-------------------------------------------------------------------------------
-- ElvUI_ChatTweaks Created By Lockslap (US, Bleeding Hollow)
-- Continued By Klix (EU, Twisting Nether)
-- Based on functionality provided by Prat and/or Chatter
-------------------------------------------------------------------------------
local Module = ElvUI_ChatTweaks:NewModule("Chat Link Icons", "AceConsole-3.0", "AceEvent-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("ElvUI_ChatTweaks", false)
Module.name = L["Chat Link Icons"]..ElvUI_ChatTweaks.NewSign
Module.namespace = string.gsub(Module.name, " ", "")

local db, options
local defaults = {
	global = {
		Links = {
			achievement = true,
			item = true,
			player = true,
			spell = true,
		},
		Icons = {
			Race = true,
			Class = true,
		},
		PawnIntegration = true,
	}
}

local print; do--	Debug use only (Bad things happen if the unmodified print() is used in our hook)
	local ScrollingMessageFrameMixin_AddMessage=ScrollingMessageFrameMixin.AddMessage;
	function print(...)
		local str="";
		for i=1,select("#",...) do str=str..(i>1 and " " or "")..tostring(select(i,...)); end
		ScrollingMessageFrameMixin_AddMessage(DEFAULT_CHAT_FRAME,str);
	end
end

--------------------------
--[[	Other Locals	]]
--------------------------
local function GenerateFromAtlas(atlas)
	local tex,w,h,x1,x2,y1,y2=GetAtlasInfo(atlas);
	if x1==0 and x2==1 and y1==0 and y2==1 then
		return ("|T%s:0|t"):format(tostring(tex))
	else
		local filew,fileh=math.floor(w/math.abs(x2-x1)+0.5),math.floor(h/math.abs(y2-y1)+0.5)
		return ("|T%s:0:0:0:0:%d:%d:%d:%d:%d:%d|t"):format(tostring(tex),filew,fileh,math.floor(x1*filew+0.5),math.floor(x2*filew+0.5),math.floor(y1*fileh+0.5),math.floor(y2*fileh+0.5))
	end
end

local LinkPattern = "|H([^:|]+):([^|]-)|h(.-)|h";
local Signature = "|c00434c49|r";--	Hex code for "CLI" as our signature (Lets us identify strings we processed already with a hidden tag)

local PawnIsLoaded = IsAddOnLoaded("Pawn");
local PawnUpgradeIcon = GenerateFromAtlas("bags-greenarrow");

local OptionTypeLookup = {
	currency = "item";
	enchant = "spell";
	trade = "spell";
	transmogappearance = "item";
	BNplayer = "player";
	unit = "player";
};

--------------------------
--[[	Texture Tables	]]
--------------------------
--	|Tpath:size1:size2:xoffset:yoffset:dimx:dimy:coordx1:coordx2:coordy1:coordy2|t
local Races = {--		Races/Genders
	Human2		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:0:64:0:64|t";
	Human3		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:0:64:128:192|t";
	Tauren2		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:0:64:64:128|t";
	Tauren3		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:0:64:192:256|t";
	Dwarf2		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:64:128:0:64|t";
	Dwarf3		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:64:128:128:192|t";
	Scourge2	="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:64:128:64:128|t";
	Scourge3	="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:64:128:192:256|t";
	Gnome2		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:128:192:0:64|t";
	Gnome3		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:128:192:128:192|t";
	Troll2		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:128:192:64:128|t";
	Troll3		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:128:192:192:256|t";
	NightElf2	="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:192:256:0:64|t";
	NightElf3	="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:192:256:128:192|t";
	Orc2		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:192:256:64:128|t";
	Orc3		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:192:256:192:256|t";
	Draenei2	="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:256:320:0:64|t";
	Draenei3	="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:256:320:128:192|t";
	BloodElf2	="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:256:320:64:128|t";
	BloodElf3	="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:256:320:192:256|t";
	Worgen2		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:320:384:0:64|t";
	Worgen3		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:320:384:128:192|t";
	Goblin2		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:320:384:64:128|t";
	Goblin3		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:320:384:192:256|t";
	Pandaren2	="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:384:448:0:64|t";
	Pandaren3	="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:384:448:128:192|t";

--[[	Future entries for Alliance/Horde differentiation if we ever get the ability
	Pandaren2Alliance	="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:384:448:0:64|t";
	Pandaren3Alliance	="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:384:448:128:192|t";
	Pandaren2Horde		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:384:448:64:128|t";
	Pandaren3Horde		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Races:0:0:0:0:512:256:384:448:192:256|t";
--]]
};
local Classes = {--		Classes
	WARRIOR		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes:0:0:0:0:256:256:0:64:0:64|t";
	MAGE		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes:0:0:0:0:256:256:64:128:0:64|t";
	ROGUE		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes:0:0:0:0:256:256:128:192:0:64|t";
	DRUID		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes:0:0:0:0:256:256:192:256:0:64|t";
	HUNTER		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes:0:0:0:0:256:256:0:64:64:128|t";
	SHAMAN		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes:0:0:0:0:256:256:64:128:64:128|t";
	PRIEST		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes:0:0:0:0:256:256:128:192:64:128|t";
	WARLOCK		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes:0:0:0:0:256:256:192:256:64:128|t";
	PALADIN		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes:0:0:0:0:256:256:0:64:128:192|t";
	DEATHKNIGHT	="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes:0:0:0:0:256:256:64:128:128:192|t";
	MONK		="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes:0:0:0:0:256:256:128:192:128:192|t";
	DEMONHUNTER	="|TInterface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes:0:0:0:0:256:256:192:256:128:192|t";
};

--------------------------
--[[	Link Transform	]]
--------------------------
local PlayerCache = {};
local LinkCache = {};

local function IsItemUpgrade(link)--	Modified PawnIsItemIDAnUpgrade() (allows full links to compare upgraded items)
	if not (PawnIsLoaded and PawnIsReady() and link) then return nil; end
	local item=PawnGetItemData(link);
	if not item then return nil; end
	return PawnIsItemAnUpgrade(item) and true or false;
end

local function IconsFromGUID(guid)
	if not guid then return "",""; end
	local _,class,_,race,gender=GetPlayerInfoByGUID(guid);
	return Races[race..gender] or "",class and Classes[class] or "";
end

--	Texture generation functions
local TextureFunctions = {
	achievement=function(id,text) return "|T"..select(10,GetAchievementInfo(tonumber(id:match("%d+"))))..":0|t"; end;
	currency=function(id,text) return "|T"..select(3,GetCurrencyInfo(tonumber(id:match("%d+"))))..":0|t"; end;
	item=function(id,text) return "|T"..GetItemIcon(tonumber(id:match("%d+")))..":0|t"; end;
	player=function(id,text) local race,class=IconsFromGUID(PlayerCache[id:match("[^:]+")]); return (db.Icons.Race and race or "")..(db.Icons.Class and class or ""); end;
	spell=function(id,text) return "|T"..select(3,GetSpellInfo(tonumber(id:match("%d+"))))..":0|t"; end;
	trade=function(id,text) return "|T"..select(3,GetSpellInfo(tonumber(id:match(":(%d+)"))))..":0|t"; end;
	transmogappearance=function(id,text) return "|T"..GetItemIcon(tonumber(select(6,C_TransmogCollection.GetAppearanceSourceInfo(tonumber(id))):match("|Hitem:(%d+)")))..":0|t"; end;
	unit=function(id,text) local race,class=IconsFromGUID(id:match("[^:]+")); return (db.Icons.Race and race or "")..(db.Icons.Class and class or ""); end;
};

--	Shared functions
TextureFunctions["enchant"] = TextureFunctions["spell"];

local function ConvertLink(link,...)--	Conversion Function
	local type,id,text=...;
	if select("#",...)<3 then type,id,text=link:match(LinkPattern); end
	local func=TextureFunctions[type];

--	Pawn Itegration (Don't cache this)
	local upgradeicon=(type=="item" and db.PawnIntegration and IsItemUpgrade("item:"..id)) and PawnUpgradeIcon or "";

--	Return if no function or link type is disabled
	if not func or db.Links[OptionTypeLookup[type] or type]==false then return link..upgradeicon; end

--	Use cached links
	if LinkCache[link] then return LinkCache[link]..upgradeicon; end

--	Rarely, the game doesn't give us the icons we need in the time we need them
--	Leting any error halt conversion and return
	local ok,pre,post=pcall(func,id,text);
	if not ok then return link..upgradeicon; end

--	Add texture to cache and return modified link
	local newlink=(pre or "")..link..(post or "");
	LinkCache[link]=newlink;
	return newlink..upgradeicon;
end

function Module:RecompileLinks(type)--	Recompilation Function
	for i,j in pairs(LinkCache) do
		if type==nil or i:match(LinkPattern)==type then
--			Need to invalidate the entry so ConvertChatLinkHelper() doesn't return our old link
--			Setting to false will fool the check and still keep the entry in the table
			LinkCache[i]=false;
			ConvertLink(i,i:match(LinkPattern));
		end
	end
end

--------------------------
--[[	Message Hook	]]
--------------------------
local ScrollingMessageFrameMixin_AddMessage = ScrollingMessageFrameMixin.AddMessage;
local function AddMessage(self,msg,...)
	if type(msg)=="string" and msg:sub(-#Signature,-1)~=Signature then
		local fix;msg,fix=msg:gsub("("..LinkPattern..")",ConvertLink);
		if fix>0 then
			msg=msg..Signature;--	Append our signature to track if this message was handled before
			while fix>0 do msg,fix=msg:gsub("(|c%x%x%x%x%x%x%x%x)(%s*|T.-|t%s*)","%2%1"); end--	Fix some links relinking with textures and blocking the ability to send them
		end
	end
	return ScrollingMessageFrameMixin_AddMessage(self,msg,...);
end;

--------------------------
--[[	Event Hooks	]]
--------------------------
--	The following code pre-hooks all chat frames to gather needed player info before each frame displays it

local function EventHandler(self,event,...)
	if event:find("^CHAT_MSG_") then
		if ChatTypeInfo[event:sub(10):gsub("^WHISPER.*","WHISPER"):gsub("^CHANNEL.*","CHANNEL"..(select(8,...) or ""))] then
--			WoW has been known to drop server names from the sender shortly after loading an area
--			This updates the stored GUID anyway in case this causes corruption in the cache
			local _,name,_,_,_,_,_,_,_,_,_,guid=...;
			if (guid or "")~="" then PlayerCache[name]=guid; end
		end

--[[		No BNPlayer support yet
		if ctype=="BN_INLINE_TOAST_ALERT" and (...)=="FRIEND_ONLINE" then
--			We need to recomplile BNPlayer links in case they log in on a new char
			AddOn.RecompileLinks("BNplayer");
		end
--]]
	end
end

local HookedFrames = {};
local function HookChatFrameEvents(frame)
	if frame and not HookedFrames[frame] then
		local func=frame:GetScript("OnEvent");
		frame:SetScript("OnEvent",func and function(...) EventHandler(...); func(...); end or EventHandler);
		frame.AddMessage = AddMessage;
		HookedFrames[frame]=true;
	end
	return frame;--	Enables passthrough calling
end

for i=1,NUM_CHAT_WINDOWS do HookChatFrameEvents(_G["ChatFrame"..i]); end
local FCF_OpenTemporaryWindow = FCF_OpenTemporaryWindow;
function _G.FCF_OpenTemporaryWindow(...) return HookChatFrameEvents(FCF_OpenTemporaryWindow(...)); end;

function Module:OnEnable()

end

function Module:OnInitialize()
	self.db = ElvUI_ChatTweaks.db:RegisterNamespace(Module.namespace, defaults)
	db = self.db.global
	self.debug = ElvUI_ChatTweaks.db.global.debugging
end

function Module:OnDisable()
    db.Links.achievement = false
	db.Links.item = false
	db.Links.player = false
	db.Links.spell = false
	db.Icons.Race = false
	db.Icons.Class = false
end

function Module:GetOptions()
	if not options then
		options = {
			achievement = {
				order = 15,
				type = "toggle",
				name = L["Achievement"],
				desc = L["Add icons to linked achievements in chat."],
				get = function(info) return db.Links[ info[#info] ] end,
				set = function(info, value) db.Links[ info[#info] ] = value; end,
			},
			item = {
				order = 17,
				type = "toggle",
				name = L["Item"],
				desc = L["Add icons to linked items in chat."],
				get = function(info) return db.Links[ info[#info] ] end,
				set = function(info, value) db.Links[ info[#info] ] = value; end,
			},
			player = {
				order = 18,
				type = "toggle",
				name = L["Player"],
				desc = L["Add icons to players in chat."],
				get = function(info) return db.Links[ info[#info] ] end,
				set = function(info, value) db.Links[ info[#info] ] = value; end,
			},
			spell = {
				order = 19,
				type = "toggle",
				name = L["Spell"],
				desc = L["Add icons to linked spells in chat."],
				get = function(info) return db.Links[ info[#info] ] end,
				set = function(info, value) db.Links[ info[#info] ] = value; end,
			},
			Race = {
				order = 20,
				type = "toggle",
				name = L["Race/Gender"],
				desc = L["Add icons to player race and gender in chat."],
				get = function(info) return db.Icons[ info[#info] ] end,
				set = function(info, value) db.Icons[ info[#info] ] = value; end,
			},
			Class = {
				order = 21,
				type = "toggle",
				name = L["Class"],
				desc = L["Add icons to player class in chat."],
				get = function(info) return db.Icons[ info[#info] ] end,
				set = function(info, value) db.Icons[ info[#info] ] = value; end,
			},
			PawnIntegration = {
				order = 21,
				type = "toggle",
				name = L["Pawn Integration"],
				desc = L["Add support for Pawn."],
				disabled = function() return not IsAddOnLoaded("Pawn") end,
				get = function(info) return db[ info[#info] ] end,
				set = function(info, value) db[ info[#info] ] = value; end,
			},
		}
	end
	return options
end

function Module:Info()
	return L["Adds icons to various links in chat.\nManually enable your prefered settings, one by one, if you disable the whole the module."]
end